import json
import time
from pathlib import Path

import click

from hashtrack import cli
from hashtrack.utils.constants import CACHE_PATH, CONFIG_PATH, FileStatus
from hashtrack.utils.log import fmt_log
from hashtrack.utils.misc import (
    abort_if_cache_not_initialized,
    load_config,
    load_cache,
    rglob,
    write_cache_,
    get_file_status
)


@cli.command()
@click.option("--strict", is_flag=True, default=False,
              help="Only allows adding new entries, not modifying existing ones.")
def watch(strict):
    """Scan for new/modified/removed files and update the cache continuously."""
    print(f"Watching for changes{' (strict)' if strict else ''}...")
    abort_if_cache_not_initialized()
    cfg = load_config(CONFIG_PATH)

    cache = load_cache(CACHE_PATH)

    search_dirs = set([Path(d) for d in cfg.search_dirs])

    cache_modified = False

    while True:
        file_paths = []
        for d in search_dirs:
            for f in rglob(str(d), cfg.extensions):
                file_paths.append(f)

                file_status, md5 = get_file_status(cache, f)
                if file_status == FileStatus.UNCHANGED:
                    continue
                if strict and file_status != FileStatus.NEW:  # only allow adding new entries
                    fmt_log(file_status, md5, str(f), skipped_update=True)
                    continue

                if file_status == FileStatus.CORRUPTED:
                    fmt_log(file_status, md5, str(f), skipped_update=True)
                    continue
                elif (
                        (file_status == FileStatus.MODIFIED or file_status == FileStatus.NEW)
                        and time.time() - f.stat().st_mtime > cfg.watch.min_update_interval
                ):
                    # Update cache
                    write_cache_(cache, rel_path=f, md5=md5)
                    cache_modified = True
                    fmt_log(file_status, md5, str(f))

        remaining_files = set(cache.keys()) - set([str(f) for f in file_paths])
        for f in remaining_files:
            file_status, md5 = FileStatus.REMOVED, "-" * 32
            if strict:
                fmt_log(file_status, md5, f, skipped_update=True)
                continue

            fmt_log(file_status, md5, f)
            del cache[str(f)]
            cache_modified = True

        if cache_modified:
            with open(CACHE_PATH, "w") as f:
                json.dump(cache, f, indent=4)
            cache_modified = False
            print("Cache updated.")

        time.sleep(cfg.watch.sleep_interval)



import shutil
from pathlib import Path

import click
import yaml

from hashtrack import cli
from hashtrack.utils.constants import TRACK_DIR, CONFIG_PATH
from hashtrack.utils.misc import abort_if_cache_not_initialized, load_config


@cli.group()
def config():
    """Configuration commands"""
    ...


@config.command()
def reset():
    """Reset the configuration"""
    abort_if_cache_not_initialized()
    shutil.copy(Path(__file__).parent.parent / "config.yml", TRACK_DIR / "config.yml")
    print("Configuration reset to default.")


@config.command()
@click.option("--extensions", type=str, default="", help="Comma-separated list of file extensions to track.")
@click.option("--search_dirs", type=str, default="", help="Comma-separated list of directories to search for files.")
def add(extensions, search_dirs):
    """Add a new configuration entry"""
    abort_if_cache_not_initialized()
    cfg = load_config(CONFIG_PATH)

    if extensions:
        for ext in extensions.split(","):
            if ext.strip() not in cfg.extensions:
                cfg.extensions.append(ext.strip())

    if search_dirs:
        for d in search_dirs.split(","):
            if d.strip() not in cfg.search_dirs:
                cfg.search_dirs.append(d.strip())

    with open(CONFIG_PATH, "w") as f:
        yaml.safe_dump(cfg.to_dict(), f, default_flow_style=False)

    print("Configuration updated.")


@config.command()
@click.option("--extensions", type=str, default="", help="Comma-separated list of file extensions to track.")
@click.option("--search_dirs", type=str, default="", help="Comma-separated list of directories to search for files.")
def remove(extensions, search_dirs):
    """Remove a configuration entry"""
    abort_if_cache_not_initialized()
    cfg = load_config(CONFIG_PATH)

    if extensions:
        for ext in extensions.split(","):
            if ext.strip() in cfg.extensions:
                cfg.extensions.remove(ext.strip())

    if search_dirs:
        for d in search_dirs.split(","):
            if d.strip() in cfg.search_dirs:
                cfg.search_dirs.remove(d.strip())

    with open(CONFIG_PATH, "w") as f:
        yaml.safe_dump(cfg.to_dict(), f, default_flow_style=False)

    print("Configuration updated.")


@config.command()
def show():
    """Show the current configuration"""
    abort_if_cache_not_initialized()
    cfg = load_config(CONFIG_PATH)
    print(yaml.safe_dump(cfg.to_dict(), default_flow_style=False))



import json

from hashtrack import cli
from hashtrack.utils.constants import CACHE_PATH
from hashtrack.utils.misc import abort_if_cache_not_initialized, load_cache


@cli.group()
def cache():
    """Operations on the cache."""
    ...


@cache.command()
def clear():
    """Clear all entries in the cache."""
    abort_if_cache_not_initialized()

    cache = load_cache(CACHE_PATH)

    key = input(f"Are you sure you want to clear {len(cache)} entries in the cache? [y/n]: ")
    if key.lower() != "y":
        print("Aborted.")
        return

    with open(CACHE_PATH, "w") as f:
        json.dump({}, f)

    print("Cache cleared.")


@cache.command()
def ls():
    """List all entries in the cache."""
    abort_if_cache_not_initialized()

    cache = load_cache(CACHE_PATH)

    for k, v in cache.items():
        print(f"{k} -> {v}")



import hashlib
from collections import defaultdict
from pathlib import Path

import click

from hashtrack import cli
from hashtrack.utils.constants import CACHE_PATH, FileStatus
from hashtrack.utils.log import fmt_log
from hashtrack.utils.misc import abort_if_cache_not_initialized, load_cache, get_file_status


@cli.command()
@click.option("-f", "--file", type=str, default="", help="File to check.")
@click.option("-v", "--verbose", is_flag=True, default=False, help="Print verbose output.")
def check(file, verbose):
    """Check if entries in cache is modified (or corrupted)"""
    abort_if_cache_not_initialized()

    cache = load_cache(CACHE_PATH)

    if file:
        file = Path(file)
        if file.is_file():
            if file.is_absolute():  # if is_absolute, convert to relative path
                file = file.relative_to(Path.cwd())

            file_status, md5 = get_file_status(cache, file)
            fmt_log(file_status, md5, str(file))
        else:
            print("You must provide a valid file path (not a directory).")

        return

    counter = defaultdict(int)

    for k, v in cache.items():
        file_status, md5 = get_file_status(cache, Path(k))
        counter[file_status] += 1

        if not verbose and file_status == FileStatus.UNCHANGED:
            continue
        fmt_log(file_status, md5, k)

    print(f"\nSummary:\n"
          f"Unchanged: {counter[FileStatus.UNCHANGED]}\n"
          f"New: {counter[FileStatus.NEW]}\n"
          f"Modified: {counter[FileStatus.MODIFIED]}\n"
          f"Removed: {counter[FileStatus.REMOVED]}\n"
          f"Corrupted: {counter[FileStatus.CORRUPTED]}")



import sys

from loguru import logger

from hashtrack.utils.constants import FileStatus

level_unchanged = logger.level("UNCHANGED", no=20, color="<green>")
level_new = logger.level("NEW", no=20, color="<blue>")
level_modified = logger.level("MODIFIED", no=20, color="<cyan>")
level_removed = logger.level("REMOVED", no=20, color="<yellow>")
level_corrupted = logger.level("CORRUPTED", no=20, color="<red>")

logger.remove()

logger.add(sys.stdout, colorize=True, format="[<level>{level:^11}</level>] {message}")
logger = logger.opt(ansi=True)

_map = {
    FileStatus.UNCHANGED: "UNCHANGED",
    FileStatus.NEW: "NEW",
    FileStatus.MODIFIED: "MODIFIED",
    FileStatus.REMOVED: "REMOVED",
    FileStatus.CORRUPTED: "CORRUPTED",
}


def log_file_status(file_status, message):
    logger.log(_map[file_status], message)


def make_fmt_log():
    """
    Make a closure for logging file status.
    """
    first_call = True

    def inner(file_status: FileStatus, md5: str, file_path: str, skipped_update=False):
        nonlocal first_call
        # [status] [md5:keep 5 chars] [file_path]
        if first_call:
            first_call = False
            print(f"{'Status':<13} {'MD5':<5} {'File Path':<50}")

        log_file_status(file_status,
                        f"{md5[:5]} {file_path: <50}{'<yellow>(skipped)</yellow>' if skipped_update else ''}")

    return inner


fmt_log = make_fmt_log()



import hashlib
import json
import os
import time
from pathlib import Path
from typing import Tuple

import yaml
from addict import Dict

from hashtrack.utils.constants import TRACK_DIR, FileStatus


def get_file_status(cache, f) -> Tuple[FileStatus, str]:
    if isinstance(f, str):
        f = Path(f)

    if not f.exists():
        return FileStatus.REMOVED, "-" * 32

    md5 = hashlib.md5(f.read_bytes()).hexdigest()

    if not str(f) in cache:  # new file
        return FileStatus.NEW, md5
    else:  # existing file
        if cache[str(f)]["md5"] != md5:
            if cache[str(f)]["modified"] < f.stat().st_mtime:
                return FileStatus.MODIFIED, md5
            return FileStatus.CORRUPTED, md5

        return FileStatus.UNCHANGED, md5


def abort_if_cache_not_initialized():
    """Check if the cache is initialized"""
    if not Path(TRACK_DIR).exists():
        print("hashtrack cache not initialized. Run `hashtrack init`.")
        exit(1)


def rglob(directory, extensions):
    for root, dirs, files in os.walk(directory):
        for file in files:
            if f".{file.split('.')[-1]}" in extensions:
                yield Path(root) / file


def load_config(config_path: Path) -> Dict:
    try:
        with open(config_path, "r") as f:
            config = yaml.safe_load(f)
        config = Dict(config)
    except Exception as e:
        print(f"Error loading config with exception: {e}. \n"
              f"Try to reset to default config by `hashtrack config reset`.")
        exit(1)

    if "." in config.search_dirs:
        print("------------------ WARNING ------------------\n"
              "Detected '.' in search_dirs.\n"
              "This is the default behavior. However, it is recommended to specify the search directories (checkpoint directories) explicitly to increase efficiency.\n"
              "Example:\n"
              "\t`hashtrack config add --search_dirs=your_checkpoint_dir`\n"
              "\t`hashtrack config remove --search_dirs='.'`\n"
              "---------------------------------------------\n")

    return config


def load_cache(cache_path: Path) -> Dict:
    with open(cache_path, "r") as f:
        cache = json.load(f)
    return cache


def write_cache_(cache, *, rel_path, md5):
    if isinstance(rel_path, str):
        rel_path = Path(rel_path)

    cache[str(rel_path)] = {
        "name": rel_path.name,
        "modified": rel_path.stat().st_mtime,
        "md5": md5,
        "cache_updated_at": time.time()
    }

    return cache



import json
import shutil
from pathlib import Path

from hashtrack import cli
from hashtrack.utils.constants import TRACK_DIR, CACHE_PATH, CONFIG_DEFAULT_PATH, CONFIG_PATH


@cli.command()
def init():
    """Initialize hashtrack cache"""
    if Path(TRACK_DIR).exists():
        print("hashtrack cache already exists.")
        return

    # Initialize TRACK_DIR
    Path(TRACK_DIR).mkdir(parents=True)

    # Copy default config.yml
    shutil.copy(CONFIG_DEFAULT_PATH, CONFIG_PATH)

    # Initialize empty cache
    with open(CACHE_PATH, "w") as f:
        json.dump({}, f)

    # if .gitignore exists, append TRACK_DIR
    gitignore_path = Path(".gitignore")
    if gitignore_path.exists():
        with open(gitignore_path, "a") as f:
            f.write(f"\n{TRACK_DIR}\n")

    print("hashtrack cache initialized.")



import json
from pathlib import Path

import click

from hashtrack import cli
from hashtrack.utils.constants import CACHE_PATH, CONFIG_PATH, FileStatus
from hashtrack.utils.log import fmt_log
from hashtrack.utils.misc import (
    abort_if_cache_not_initialized,
    load_config,
    load_cache,
    rglob,
    write_cache_,
    get_file_status
)


@cli.command()
@click.option("--strict", is_flag=True, default=False,
              help="Only allows adding new entries, not modifying existing ones.")
@click.option("-y", is_flag=True, default=False, help="Skip confirmation prompt.")
def update(strict, y):
    """Scan for new/modified/removed files and update the cache."""
    print(f"Updating cache{' (strict)' if strict else ''}...")

    abort_if_cache_not_initialized()
    cfg = load_config(CONFIG_PATH)

    cache = load_cache(CACHE_PATH)

    search_dirs = set([Path(d) for d in cfg.search_dirs])

    cache_modified = False

    file_paths = []
    for d in search_dirs:
        for f in rglob(str(d), cfg.extensions):
            file_paths.append(f)

            file_status, md5 = get_file_status(cache, f)
            if file_status == FileStatus.UNCHANGED:
                continue
            if strict and file_status != FileStatus.NEW:  # only allow adding new entries
                fmt_log(file_status, md5, str(f), skipped_update=True)
                continue

            if file_status != FileStatus.CORRUPTED:
                # Update cache
                write_cache_(cache, rel_path=f, md5=md5)
                cache_modified = True
                fmt_log(file_status, md5, str(f))
            else:
                fmt_log(file_status, md5, str(f), skipped_update=True)

    remaining_files = set(cache.keys()) - set([str(f) for f in file_paths])
    for f in remaining_files:
        file_status, md5 = FileStatus.REMOVED, "-" * 32
        if strict:
            fmt_log(file_status, md5, f, skipped_update=True)
            continue

        fmt_log(file_status, md5, f)
        del cache[str(f)]
        cache_modified = True

    if not cache_modified:
        print("Cache is up-to-date.")
        return

    if y:
        key = "y"
    else:
        key = input("Update cache? [y/n]: ")

    if key.lower() == "y":
        with open(CACHE_PATH, "w") as f:
            json.dump(cache, f, indent=4)
        print("Cache updated.")
    else:
        print("Aborted.")



from enum import Enum, auto
from pathlib import Path


class FileStatus(Enum):
    UNCHANGED = auto()
    NEW = auto()
    MODIFIED = auto()
    REMOVED = auto()
    CORRUPTED = auto()


TRACK_DIR = Path(".hashtrack")
CACHE_PATH = TRACK_DIR / "cache.json"
CONFIG_DEFAULT_PATH = Path(__file__).parent.parent / "config.yml"
CONFIG_PATH = TRACK_DIR / "config.yml"
